"""Utility classes and functions for configuring and setting up the content and the look of a QR code."""
import decimal
from collections import namedtuple
from django.utils.html import escape
from qr_code.qrcode.constants import DEFAULT_MODULE_SIZE, SIZE_DICT, \
    DEFAULT_ERROR_CORRECTION, DEFAULT_IMAGE_FORMAT


class QRCodeOptions:
    """
    Represents the options used to create and draw a QR code.
    """
    def __init__(self, *, size=DEFAULT_MODULE_SIZE, border=4, version=None,
                 image_format='svg', error_correction=DEFAULT_ERROR_CORRECTION,
                 micro=False, dark_color='#000', light_color='#fff',
                 finder_dark_color=False, finder_light_color=False,
                 data_dark_color=False, data_light_color=False,
                 version_dark_color=False, version_light_color=False,
                 format_dark_color=False, format_light_color=False,
                 alignment_dark_color=False, alignment_light_color=False,
                 timing_dark_color=False, timing_light_color=False,
                 separator_color=False, dark_module_color=False, 
                 quiet_zone_color=False):
        """
        :param size: The size of the QR code as an integer or a string.
                    Default is *'m'*.
        :type: str or int
        :param int border: The size of the border (blank space around the code).
        :param int version: The version of the QR code gives the size of the matrix.
                Default is *None* which mean automatic in order to avoid data overflow.
        :param str image_format: The graphics format used to render the QR code.
                It can be either *'svg'* or *'png'*. Default is *'svg'*.
        :param str error_correction: How much error correction that might be required
                to read the code. It can be either *'L'*, *'M'*, *'Q'*, or *'H'*. Default is *'M'*.
        :param bool micro: Indicates if a Micro QR Code should be created. Default: False
        :param dark_color: Color of the dark modules (default: black). The
                color can be provided as ``(R, G, B)`` tuple, as hexadecimal
                format (``#RGB``, ``#RRGGBB`` ``RRGGBBAA``), or web color
                name (i.e. ``red``).
        :param light_color: Color of the light modules (default: white).
                See `color` for valid values. If light is set to ``None`` the
                light modules will be transparent.
        :param finder_dark_color: Color of the dark finder modules (default: same as ``dark_color``)
        :param finder_light_color: Color of the light finder modules (default: same as ``light_color``)
        :param data_dark_color: Color of the dark data modules (default: same as ``dark_color``)
        :param data_light_color: Color of the light data modules (default: same as ``light_color``)
        :param version_dark_color: Color of the dark version modules (default: same as ``dark_color``)
        :param version_light_color: Color of the light version modules (default: same as ``light_color``)
        :param format_dark_color: Color of the dark format modules (default: same as ``dark_color``)
        :param format_light_color: Color of the light format modules (default: same as ``light_color``)
        :param alignment_dark_color: Color of the dark alignment modules (default: same as ``dark_color``)
        :param alignment_light_color: Color of the light alignment modules (default: same as ``light_color``)
        :param timing_dark_color: Color of the dark timing pattern modules (default: same as ``dark_color``)
        :param timing_light_color: Color of the light timing pattern modules (default: same as ``light_color``)
        :param separator_color: Color of the separator (default: same as ``light_color``)
        :param dark_module_color: Color of the dark module (default: same as ``dark_color``)
        :param quiet_zone_color: Color of the quiet zone modules (default: same as ``light_color``)

        The *size* parameter gives the size of each module of the QR code matrix. It can be either a positive integer or one of the following letters:
            * t or T: tiny (value: 6)
            * s or S: small (value: 12)
            * m or M: medium (value: 18)
            * l or L: large (value: 30)
            * h or H: huge (value: 48)

        For PNG image format the size unit is in pixels, while the unit is 0.1 mm for SVG format.

        The *border* parameter controls how many modules thick the border should be (blank space around the code).
        The default is 4, which is the minimum according to the specs.

        The *version* parameter is an integer from 1 to 40 that controls the size of the QR code matrix. Set to None to determine
        this automatically. The smallest, version 1, is a 21 x 21 matrix. The biggest, version 40, is 177 x 177 matrix.
        The size grows by 4 modules/side.

        There are 4 error correction levels used for QR codes, with each one adding different amounts of "backup" data
        depending on how much damage the QR code is expected to suffer in its intended environment, and hence how much
        error correction may be required. The correction level can be configured with the *error_correction* parameter as follow:
            * l or L: error correction level L – up to 7% damage
            * m or M: error correction level M – up to 15% damage
            * q or Q: error correction level Q – up to 25% damage
            * h or H: error correction level H – up to 30% damage
        :raises: TypeError in case an unknown argument is given.
        """
        self._size = size
        self._border = border
        if _can_be_cast_to_int(version):
            version = int(version)
            if not 1 <= version <= 40:
                version = None
        elif version in ('m1', 'm2', 'm3', 'm4', 'M1', 'M2', 'M3', 'M4'):
            version = version.lower()
            # Set / change the micro setting otherwise Segno complains about
            # conflicting parameters
            micro = True
        else:
            version = None
        self._version = version
        if not isinstance(micro, bool):
            micro = micro == 'True'
        self._micro = micro
        try:
            error = error_correction.lower()
            self._error_correction = error if error in ('l', 'm', 'q', 'h') else DEFAULT_ERROR_CORRECTION
        except AttributeError:
            self._error_correction = DEFAULT_ERROR_CORRECTION
        try:
            image_format = image_format.lower()
            self._image_format = image_format if image_format in ('svg', 'png') else DEFAULT_IMAGE_FORMAT
        except AttributeError:
            self._image_format = DEFAULT_IMAGE_FORMAT
        self._colors = dict(dark_color=dark_color, light_color=light_color,
                            finder_dark_color=finder_dark_color,
                            finder_light_color=finder_light_color,
                            data_dark_color=data_dark_color,
                            data_light_color=data_light_color,
                            version_dark_color=version_dark_color,
                            version_light_color=version_light_color,
                            format_dark_color=format_dark_color,
                            format_light_color=format_light_color,
                            alignment_dark_color=alignment_dark_color,
                            alignment_light_color=alignment_light_color,
                            timing_dark_color=timing_dark_color,
                            timing_light_color=timing_light_color,
                            separator_color=separator_color,
                            dark_module_color=dark_module_color,
                            quiet_zone_color=quiet_zone_color)

    def kw_make(self):
        """Internal method which returns a dict of parameters to create a QR code.

        :rtype: dict
        """
        return dict(version=self._version, error=self._error_correction,
                    micro=self._micro)

    def kw_save(self):
        """Internal method which returns a dict of parameters to save a QR code.

        :rtype: dict
        """
        image_format = self._image_format
        kw = dict(kind=image_format, scale=self._size_as_int())
        # Change the color mapping into the keywords Segno expects
        # (remove the "_color" suffix from the module names)
        kw.update({k[:-6]: v for k, v in self.color_mapping().items()})
        if image_format == 'svg':
            kw['unit'] = 'mm'
            scale = decimal.Decimal(kw['scale']) / 10
            kw['scale'] = scale
        return kw

    def color_mapping(self):
        """Internal method which returns the color mapping.

        Only non-default values are returned.

        :rtype: dict
        """
        colors = {k: v for k, v in self._colors.items() if v is not False}
        # Remove common default "dark" and "light" values
        if colors.get('dark_color') in ('#000', '#000000', 'black'):
            del colors['dark_color']
        if colors.get('light_color') in ('#fff', '#FFF', '#ffffff', '#FFFFFF', 'white'):
            del colors['light_color']
        return colors

    def _size_as_int(self):
        """Returns the size as integer value.

        :rtype: int
        """
        size = self._size
        if _can_be_cast_to_int(size):
            actual_size = int(size)
            if actual_size < 1:
                actual_size = SIZE_DICT[DEFAULT_MODULE_SIZE]
        elif isinstance(size, str):
            actual_size = SIZE_DICT.get(size.lower(), DEFAULT_MODULE_SIZE)
        else:
            actual_size = SIZE_DICT[DEFAULT_MODULE_SIZE]
        return actual_size

    @property
    def size(self):
        return self._size

    @property
    def border(self):
        return self._border

    @property
    def version(self):
        return self._version

    @property
    def image_format(self):
        return self._image_format

    @property
    def error_correction(self):
        return self._error_correction

    @property
    def micro(self):
        return self._micro


def _can_be_cast_to_int(value):
    return isinstance(value, int) or (isinstance(value, str) and value.isdigit())


class ContactDetail:
    """
    Represents the detail of a contact.

    The following fields are provided:
        * first_name
        * last_name
        * first_name_reading: the sound of the first name.
        * last_name_reading: the sound of the last name.
        * tel: the phone number, it can appear multiple times.
        * tel_av: the video-phone number, it can appear multiple times.
        * email: the email address, it can appear multiple times.
        * memo: notes.
        * birthday: the birth date (Python date).
        * address: the fields divided by commas (,) denote PO box, room number, house number, city, prefecture, zip code and country, in order.
        * url: homepage URL.
        * nickname: display name.
        * org: organization or company name (non-standard,but often recognized, ORG field).
    """
    first_name = None
    last_name = None
    first_name_reading = None
    last_name_reading = None
    tel = None
    tel_av = None
    email = None
    memo = None
    birthday = None
    address = None
    url = None
    nickname = None
    org = None

    def __init__(self, **kwargs):
        self.first_name = kwargs.get('first_name')
        self.last_name = kwargs.get('last_name')
        self.first_name_reading = kwargs.get('first_name_reading')
        self.last_name_reading = kwargs.get('last_name_reading')
        self.tel = kwargs.get('tel')
        self.tel_av = kwargs.get('tel_av')
        self.email = kwargs.get('email')
        self.memo = kwargs.get('memo')
        self.birthday = kwargs.get('birthday')
        self.address = kwargs.get('address')
        self.url = kwargs.get('url')
        self.nickname = kwargs.get('nickname')
        self.org = kwargs.get('org')

    def make_qr_code_text(self):
        """
        Make a text for configuring a contact in a phone book. The MeCARD format is used, with an optional, non-standard (but often recognized) ORG field.

        See this archive of the format specifications: https://web.archive.org/web/20160304025131/https://www.nttdocomo.co.jp/english/service/developer/make/content/barcode/function/application/addressbook/index.html

        :return: the MeCARD contact description.
        """

        _escape_mecard_special_chars_in_object_fields(self, ('first_name', 'last_name', 'first_name_reading', 'last_name_reading', 'tel', 'tel_av', 'email', 'memo', 'url', 'nickname', 'org'))

        # See this for an archive of the format specifications:
        # https://web.archive.org/web/20160304025131/https://www.nttdocomo.co.jp/english/service/developer/make/content/barcode/function/application/addressbook/index.html
        contact_text = 'MECARD:'
        for name_components_pair in (('N:%s;', (self.escaped_last_name, self.escaped_first_name)), ('SOUND:%s;', (self.escaped_last_name_reading, self.escaped_first_name_reading))):
            if name_components_pair[1][0] and name_components_pair[1][1]:
                name = '%s,%s' % name_components_pair[1]
            else:
                name = name_components_pair[1][0] if name_components_pair[1][0] else name_components_pair[1][1]
            if name:
                contact_text += name_components_pair[0] % name
        if self.tel:
            contact_text += 'TEL:%s;' % self.escaped_tel
        if self.tel_av:
            contact_text += 'TEL-AV:%s;' % self.escaped_tel_av
        if self.email:
            contact_text += 'EMAIL:%s;' % self.escaped_email
        if self.memo:
            contact_text += 'NOTE:%s;' % self.escaped_memo
        if self.birthday:
            # Format date to YYMMDD.
            contact_text += 'BDAY:%s;' % self.birthday.strftime('%Y%m%d')
        if self.address:
            contact_text += 'ADR:%s;' % self.address
        if self.url:
            contact_text += 'URL:%s;' % self.escaped_url
        if self.nickname:
            contact_text += 'NICKNAME:%s;' % self.escaped_nickname
        # Not standard, but recognized by several readers.
        if self.org:
            contact_text += 'ORG:%s;' % self.escaped_org
        contact_text += ';'
        return contact_text


class WifiConfig(object):
    """
    Represents the configuration of a Wi-Fi connexion.

    The following fields are provided:
        * ssid: the name of the SSID
        * authentication: the authentication type for the SSID; can be AUTHENTICATION.wep or AUTHENTICATION.wpa, or AUTHENTICATION.nopass for no password. Or, omit for no password.
        * password: the password, ignored if "authentsication" is 'nopass' (in which case it may be omitted).
        * hidden: tells whether the SSID is hidden or not; can be True or False.
    """
    AUTHENTICATION = namedtuple('AUTHENTICATION', 'nopass WEP WPA')._make(range(3))
    AUTHENTICATION_CHOICES = ((AUTHENTICATION.nopass, 'nopass'), (AUTHENTICATION.WEP, 'WEP'), (AUTHENTICATION.WPA, 'WPA'))
    ssid = None
    authentication = AUTHENTICATION.nopass
    password = None
    hidden = False

    def __init__(self, **kwargs):
        self.ssid = kwargs.get('ssid')
        self.authentication = kwargs.get('authentication', WifiConfig.AUTHENTICATION.nopass)
        self.password = kwargs.get('password')
        self.hidden = kwargs.get('hidden')

    def make_qr_code_text(self):
        """
        Make a text for configuring a Wi-Fi connexion. The syntax is inspired by the MeCARD format used for contacts.

        :return: the WIFI configuration text that can be translated to a QR code.
        """

        _escape_mecard_special_chars_in_object_fields(self, ('ssid', 'password'))

        wifi_config = 'WIFI:'
        if self.ssid:
            wifi_config += 'S:%s;' % self.escaped_ssid
        if self.authentication:
            wifi_config += 'T:%s;' % WifiConfig.AUTHENTICATION_CHOICES[self.authentication][1]
        if self.password:
            wifi_config += 'P:%s;' % self.escaped_password
        if self.hidden:
            wifi_config += 'H:%s;' % str(self.hidden).lower()
        wifi_config += ';'
        return wifi_config


class Coordinates:
    def __init__(self, latitude, longitude, altitude=None):
        self.latitude = latitude
        self.longitude = longitude
        self.altitude = altitude

    def __str__(self):
        if self.altitude:
            return 'latitude: %s, longitude: %s, altitude: %s' % (self.latitude, self.longitude, self.altitude)
        return 'latitude: %s, longitude: %s' % (self.latitude, self.longitude)

    def make_geolocation_text(coordinates):
        return 'geo:%s,%s,%s' % (
            escape(coordinates.latitude), escape(coordinates.longitude), escape(coordinates.altitude))

    def make_google_maps_text(coordinates):
        return 'https://maps.google.com/local?q=%s,%s' % (
            escape(coordinates.latitude), escape(coordinates.longitude))


def make_email_text(email):
    return 'mailto:%s' % email


def make_tel_text(phone_number):
    return 'tel:%s' % phone_number


def make_sms_text(phone_number):
    return 'sms:%s' % phone_number


def make_youtube_text(video_id):
    return 'https://www.youtube.com/watch/?v=%s' % escape(video_id)


def make_google_play_text(package_id):
    return 'https://play.google.com/store/apps/details?id=%s' % escape(package_id)


def _escape_mecard_special_chars(string_to_escape):
    if not string_to_escape:
        return string_to_escape
    special_chars = ['\\', '"', ';', ',', ':']
    for sc in special_chars:
        string_to_escape = string_to_escape.replace(sc, '\\%s' % sc)
    return string_to_escape


def _escape_mecard_special_chars_in_object_fields(obj, keys):
    for key in keys:
        if hasattr(obj, key):
            setattr(obj, 'escaped_%s' % key, _escape_mecard_special_chars(getattr(obj, key)))



"""Tools for generating QR codes. This module depends on the Segno library."""
import io
from django.utils.encoding import force_str
from django.utils.html import escape
from django.utils.safestring import mark_safe
import segno
from qr_code.qrcode.constants import DEFAULT_CACHE_ENABLED, \
    DEFAULT_URL_SIGNATURE_ENABLED
from qr_code.qrcode.serve import make_qr_code_url
from qr_code.qrcode.utils import QRCodeOptions


def make_qr(text, qr_code_options):
    """Creates a QR code

    :rtype: segno.QRCode
    """
    return segno.make(force_str(text), **qr_code_options.kw_make())


def make_qr_code_image(text, qr_code_options):
    """
    Returns a bytes object representing a QR code image for the provided text.

    :param str text: The text to encode
    :param qr_code_options: Options to create and serialize the QR code.
    :rtype: bytes
    """
    qr = make_qr(text, qr_code_options)
    out = io.BytesIO()
    qr.save(out, **qr_code_options.kw_save())
    return out.getvalue()


def make_embedded_qr_code(text, qr_code_options):
    """
    Generates a <svg> or <img> tag representing the QR code for the given text.
    This tag can be embedded into an HTML document.
    """
    qr = make_qr(text, qr_code_options)
    kw = qr_code_options.kw_save()
    # Pop the image format from the keywords since qr.png_data_uri / qr.svg_inline
    # set it automatically
    kw.pop('kind')
    if qr_code_options.image_format == 'png':
        return mark_safe('<img src="{0}" alt="{1}">'
                         .format(qr.png_data_uri(**kw), escape(text)))
    return mark_safe(qr.svg_inline(**kw))


def make_qr_code_with_args(text, qr_code_args):
    options = _options_from_args(qr_code_args)
    return make_embedded_qr_code(text, options)


def make_qr_code_url_with_args(text, qr_code_args):
    cache_enabled = qr_code_args.pop('cache_enabled', DEFAULT_CACHE_ENABLED)
    url_signature_enabled = qr_code_args.pop('url_signature_enabled', DEFAULT_URL_SIGNATURE_ENABLED)
    options = _options_from_args(qr_code_args)
    return make_qr_code_url(text, options, cache_enabled=cache_enabled,
                            url_signature_enabled=url_signature_enabled)


def _options_from_args(args):
    """Returns a QRCodeOptions instance from the provided arguments.
    """
    options = args.get('options')
    if options:
        if not isinstance(options, QRCodeOptions):
            raise TypeError('The options argument must be of type QRCodeOptions.')
    else:
        # Convert the string "None" into None
        kw = {k: v if v != 'None' else None for k, v in args.items()}
        options = QRCodeOptions(**kw)
    return options



import base64
import binascii
import functools

from django.conf import settings
from django.core.exceptions import PermissionDenied, SuspiciousOperation
from django.core.signing import BadSignature, Signer
from django.http import HttpResponse
from django.views.decorators.cache import cache_page
from django.views.decorators.http import condition

from qr_code.qrcode import constants
from qr_code.qrcode.maker import make_qr_code_image
from qr_code.qrcode.utils import QRCodeOptions
from qr_code.qrcode.serve import get_url_protection_options, \
    get_qr_url_protection_token, qr_code_etag, \
    qr_code_last_modified, allows_external_request_from_user


def cache_qr_code():
    """
    Decorator that caches the requested page if a settings named 'QR_CODE_CACHE_ALIAS' exists and is not empty or None.
    """
    def decorator(view_func):
        @functools.wraps(view_func)
        def _wrapped_view(request, *view_args, **view_kwargs):
            cache_enabled = request.GET.get('cache_enabled', True)
            if cache_enabled and hasattr(settings, 'QR_CODE_CACHE_ALIAS') and settings.QR_CODE_CACHE_ALIAS:
                # We found a cache alias for storing the generate qr code and cache is enabled, use it to cache the
                # page.
                timeout = settings.CACHES[settings.QR_CODE_CACHE_ALIAS]['TIMEOUT']
                key_prefix = 'token=%s.user_pk=%s' % (request.GET.get('url_signature_enabled') or constants.DEFAULT_URL_SIGNATURE_ENABLED, request.user.pk)
                response = cache_page(timeout, cache=settings.QR_CODE_CACHE_ALIAS, key_prefix=key_prefix)(view_func)(request, *view_args, **view_kwargs)
            else:
                # No cache alias for storing the generated qr code, call the view as is.
                response = (view_func)(request, *view_args, **view_kwargs)
            return response
        return _wrapped_view
    return decorator


@condition(etag_func=qr_code_etag, last_modified_func=qr_code_last_modified)
@cache_qr_code()
def serve_qr_code_image(request):
    """Serve an image that represents the requested QR code."""
    qr_code_options = get_qr_code_option_from_request(request)
    # Handle image access protection (we do not allow external requests for anyone).
    check_image_access_permission(request, qr_code_options)
    try:
        text = base64.urlsafe_b64decode(request.GET.get('text', ''))
    except binascii.Error:
        raise SuspiciousOperation("Invalid base64 encoded string.")
    img = make_qr_code_image(text, qr_code_options=qr_code_options)
    return HttpResponse(content=img,
                        content_type='image/svg+xml' if qr_code_options.image_format == 'svg' else 'image/png')


def get_qr_code_option_from_request(request):
    request_query = request.GET.dict()
    for key in ('text', 'token', 'cache_enabled'):
        request_query.pop(key, None)
    return QRCodeOptions(**request_query)


def check_image_access_permission(request, qr_code_options):
    """Handle image access protection (we do not allow external requests for anyone)."""
    token = request.GET.get('token', '')
    if token:
        check_url_signature_token(qr_code_options, token)
    else:
        if not allows_external_request_from_user(request.user):
            raise PermissionDenied("You are not allowed to access this QR code.")


def check_url_signature_token(qr_code_options, token):
    url_protection_options = get_url_protection_options()
    signer = Signer(key=url_protection_options[constants.SIGNING_KEY],
                    salt=url_protection_options[constants.SIGNING_SALT])
    try:
        # Check signature.
        url_protection_string = signer.unsign(token)
        # Check that the given token matches the request parameters.
        random_token = url_protection_string.split('.')[-1]
        if get_qr_url_protection_token(qr_code_options, random_token) != url_protection_string:
            raise PermissionDenied("Request query does not match protection token.")
    except BadSignature:
        raise PermissionDenied("Wrong token signature.")



import re

from setuptools import setup

# Get version without importing
with open('qr_code/__init__.py', 'rb') as f:
    VERSION = str(re.search('__version__ = \'(.+?)\'', f.read().decode('utf-8')).group(1))

setup(
    name='django-qr-code',
    version=VERSION,
    packages=['qr_code', 'qr_code.qrcode', 'qr_code.templatetags'],
    url='https://github.com/dprog-philippe-docourt/django-qr-code',
    license='BSD 3-clause',
    author='Philippe Docourt',
    author_email='philippe@docourt.ch',
    maintainer='Philippe Docourt',
    description='An application that provides tools for displaying QR codes on your Django site.',
    long_description="""This application provides tools for displaying QR codes on your `Django <https://www.djangoproject.com/>`_ site.

This application depends on the `Segno QR Code generator <https://pypi.org/project/segno/>`_.

This app makes no usage of the Django models and therefore do not use any database.

Only Python >= 3.6 is supported.""",
    install_requires=['qrcode', 'django'],
    python_requires='>=3',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet :: WWW/HTTP',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3 :: Only',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Natural Language :: English'
    ],
    keywords='qr code django',
)



from datetime import datetime

QR_CODE_GENERATION_VERSION_DATE = datetime(year=2020, month=9, day=8, hour=12)
SIZE_DICT = {'t': 6, 's': 12, 'm': 18, 'l': 30, 'h': 48}
ERROR_CORRECTION_DICT = {'L': 'l', 'M': 'm', 'Q': 'q', 'H': 'h'}
DEFAULT_MODULE_SIZE = 'm'
DEFAULT_BORDER_SIZE = 4
DEFAULT_VERSION = None
DEFAULT_IMAGE_FORMAT = 'svg'
DEFAULT_ERROR_CORRECTION = 'm'
DEFAULT_CACHE_ENABLED = True
DEFAULT_URL_SIGNATURE_ENABLED = True

ALLOWS_EXTERNAL_REQUESTS_FOR_REGISTERED_USER = 'ALLOWS_EXTERNAL_REQUESTS_FOR_REGISTERED_USER'
SIGNING_KEY = 'SIGNING_KEY'
TOKEN_LENGTH = 'TOKEN_LENGTH'
SIGNING_SALT = 'SIGNING_SALT'



"""Tags for Django template system that help generating QR codes."""

from django import template

from qr_code.qrcode.maker import make_qr_code_with_args, make_qr_code_url_with_args
from qr_code.qrcode.utils import make_email_text, make_google_play_text, make_tel_text, make_sms_text, \
    make_youtube_text, WifiConfig, ContactDetail, Coordinates

register = template.Library()


def _make_contact_or_wifi_qr_code(contact_or_wifi, expected_cls, embedded, qr_code_args):
    if not isinstance(contact_or_wifi, expected_cls):
        # For compatibility with existing views and templates, try to build from dict.
        contact_or_wifi = expected_cls(**contact_or_wifi)
    if embedded:
        return make_qr_code_with_args(contact_or_wifi.make_qr_code_text(), qr_code_args=qr_code_args)
    else:
        return make_qr_code_url_with_args(contact_or_wifi.make_qr_code_text(), qr_code_args=qr_code_args)


def _make_google_maps_qr_code(embedded, **kwargs):
    if 'coordinates' in kwargs:
        coordinates = kwargs.pop('coordinates')
    else:
        coordinates = Coordinates(kwargs.pop('latitude'), kwargs.pop('longitude'))
    if embedded:
        return make_qr_code_with_args(coordinates.make_google_maps_text(), qr_code_args=kwargs)
    else:
        return make_qr_code_url_with_args(coordinates.make_google_maps_text(), qr_code_args=kwargs)


def _make_geolocation_qr_code(embedded, **kwargs):
    if 'coordinates' in kwargs:
        coordinates = kwargs.pop('coordinates')
    else:
        coordinates = Coordinates(kwargs.pop('latitude'), kwargs.pop('longitude'), kwargs.pop('altitude'))
    if embedded:
        return make_qr_code_with_args(coordinates.make_geolocation_text(), qr_code_args=kwargs)
    else:
        return make_qr_code_url_with_args(coordinates.make_geolocation_text(), qr_code_args=kwargs)


@register.simple_tag()
def qr_from_text(text, **kwargs):
    return make_qr_code_with_args(text, qr_code_args=kwargs)


@register.simple_tag()
def qr_for_email(email, **kwargs):
    return make_qr_code_with_args(make_email_text(email), qr_code_args=kwargs)


@register.simple_tag()
def qr_for_tel(phone_number, **kwargs):
    return make_qr_code_with_args(make_tel_text(phone_number), qr_code_args=kwargs)


@register.simple_tag()
def qr_for_sms(phone_number, **kwargs):
    return make_qr_code_with_args(make_sms_text(phone_number), qr_code_args=kwargs)


@register.simple_tag()
def qr_for_geolocation(**kwargs):
    """Accepts a *'coordinates'* keyword argument or a triplet *'latitude'*, *'longitude'*, and *'altitude'*."""
    return _make_geolocation_qr_code(embedded=True, **kwargs)


@register.simple_tag()
def qr_for_google_maps(**kwargs):
    """Accepts a *'coordinates'* keyword argument or a pair *'latitude'* and *'longitude'*."""
    return _make_google_maps_qr_code(embedded=True, **kwargs)


@register.simple_tag()
def qr_for_youtube(video_id, **kwargs):
    return make_qr_code_with_args(make_youtube_text(video_id), qr_code_args=kwargs)


@register.simple_tag()
def qr_for_google_play(package_id, **kwargs):
    return make_qr_code_with_args(make_google_play_text(package_id), qr_code_args=kwargs)


@register.simple_tag()
def qr_for_contact(contact_detail, **kwargs):
    return _make_contact_or_wifi_qr_code(contact_detail, ContactDetail, embedded=True, qr_code_args=kwargs)


@register.simple_tag()
def qr_for_wifi(wifi_config, **kwargs):
    return _make_contact_or_wifi_qr_code(wifi_config, WifiConfig, embedded=True, qr_code_args=kwargs)


@register.simple_tag()
def qr_url_from_text(text, **kwargs):
    return make_qr_code_url_with_args(text, qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_email(email, **kwargs):
    return make_qr_code_url_with_args(make_email_text(email), qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_tel(phone_number, **kwargs):
    return make_qr_code_url_with_args(make_tel_text(phone_number), qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_sms(phone_number, **kwargs):
    return make_qr_code_url_with_args(make_sms_text(phone_number), qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_geolocation(**kwargs):
    """Accepts a *'coordinates'* keyword argument or a triplet *'latitude'*, *'longitude'*, and *'altitude'*."""
    return _make_geolocation_qr_code(embedded=False, **kwargs)


@register.simple_tag()
def qr_url_for_google_maps(**kwargs):
    """Accepts a *'coordinates'* keyword argument or a pair *'latitude'* and *'longitude'*."""
    return _make_google_maps_qr_code(embedded=False, **kwargs)


@register.simple_tag()
def qr_url_for_youtube(video_id, **kwargs):
    return make_qr_code_url_with_args(make_youtube_text(video_id), qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_google_play(package_id, **kwargs):
    return make_qr_code_url_with_args(make_google_play_text(package_id), qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_contact(contact_detail, **kwargs):
    return _make_contact_or_wifi_qr_code(contact_detail, ContactDetail, embedded=False, qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_wifi(wifi_config, **kwargs):
    return _make_contact_or_wifi_qr_code(wifi_config, WifiConfig, embedded=False, qr_code_args=kwargs)
